/*
space - very simple space flight simulator using pitch-yaw-roll method
Copyright (C) 2006 Benjamin Schieder

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; version 2 of the License.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

*/
#include <space.h>

#define deg2rad M_PI/180
#define rad2deg 180/M_PI

CGLI_Interface	iface;

int WindW, WindH;

Vector forward, up, right, position, acceleration;
GLfloat thrust;

class CGLI_Space : public CGLI_Object { // {{{
	public:
		CGLI_Space();
		virtual ~CGLI_Space();

		virtual void Draw();

		virtual void UpdateObject();
		virtual void DoAnimationTick();

	private:
		virtual void ProcessInternalCallBack(GLuint obj_id);
};

CGLI_Space::CGLI_Space(){
}

CGLI_Space::~CGLI_Space(){
}

void CGLI_Space::Draw(){
        glBegin(GL_TRIANGLES);
                glColor3f(1.0f,0.0f,0.0f);                      // Red
                glVertex3f( 0.0f, 1.0f, 0.0f);                  // Top Of Triangle (Front)
                glColor3f(0.0f,1.0f,0.0f);                      // Green
                glVertex3f(-1.0f,-1.0f, 1.0f);                  // Left Of Triangle (Front)
                glColor3f(0.0f,0.0f,1.0f);                      // Blue
                glVertex3f( 1.0f,-1.0f, 1.0f);                  // Right Of Triangle (Front)
                glColor3f(1.0f,0.0f,0.0f);                      // Red
                glVertex3f( 0.0f, 1.0f, 0.0f);                  // Top Of Triangle (Right)
                glColor3f(0.0f,0.0f,1.0f);                      // Blue
                glVertex3f( 1.0f,-1.0f, 1.0f);                  // Left Of Triangle (Right)
                glColor3f(0.0f,1.0f,0.0f);                      // Green
                glVertex3f( 1.0f,-1.0f, -1.0f);                 // Right Of Triangle (Right)
                glColor3f(1.0f,0.0f,0.0f);                      // Red
                glVertex3f( 0.0f, 1.0f, 0.0f);                  // Top Of Triangle (Back)
                glColor3f(0.0f,1.0f,0.0f);                      // Green
                glVertex3f( 1.0f,-1.0f, -1.0f);                 // Left Of Triangle (Back)
                glColor3f(0.0f,0.0f,1.0f);                      // Blue
                glVertex3f(-1.0f,-1.0f, -1.0f);                 // Right Of Triangle (Back)
                glColor3f(1.0f,0.0f,0.0f);                      // Red
                glVertex3f( 0.0f, 1.0f, 0.0f);                  // Top Of Triangle (Left)
                glColor3f(0.0f,0.0f,1.0f);                      // Blue
                glVertex3f(-1.0f,-1.0f,-1.0f);                  // Left Of Triangle (Left)
                glColor3f(0.0f,1.0f,0.0f);                      // Green
                glVertex3f(-1.0f,-1.0f, 1.0f);                  // Right Of Triangle (Left)
        glEnd();
}

void CGLI_Space::UpdateObject(){
}

void CGLI_Space::DoAnimationTick(){
}

void CGLI_Space::ProcessInternalCallBack(GLuint obj_id){
	/* foo */
}//}}}

CGLI_Space	*object;

Vector Cross(Vector v1, Vector v2){/*{{{*/
	/* Calculates the cross product of two vectors
	 * Cross(right, up) = forward	Cross(up, right) = back
	 * Cross(forward, right) = up	Cross(right, forward) = down
	 * Cross(up, forward) = right	Cross(forward, up) = left
	 */
	Vector r;

	r.x = ((v1.y * v2.z) - (v1.z * v2.y));
	r.y = ((v1.z * v2.x) - (v1.x * v2.z));
	r.z = ((v1.x * v2.y) - (v1.y * v2.x));

	return r;
}/*}}}*/
void UpDown(Vector *forward, Vector *up, float degree){/*{{{*/
	/* here we move the ships nose up or down depending on the value of degree.
	* The only two vectors we change here are forward and up
	*/
	Vector nforward, nup, right, nright;
	float radian;
	float c, s;

	radian=degree*deg2rad;
	c=cos(radian);
	s=sin(radian);

	//right = Cross(*up, *forward);

	nforward.x = forward->x*c-up->x*s;
	nforward.y = forward->y*c-up->y*s;
	nforward.z = forward->z*c-up->z*s;

	nup.x = forward->x*s+up->x*c;
	nup.y = forward->y*s+up->y*c;
	nup.z = forward->z*s+up->z*c;

	//nright = Cross(*up, *forward);

	up->x=nup.x;
	up->y=nup.y;
	up->z=nup.z;
	forward->x=nforward.x;
	forward->y=nforward.y;
	forward->z=nforward.z;
	//right=nright;
}/*}}}*/
void LeftRight(Vector *forward, Vector *up, float degree){/*{{{*/
	/* here we move the ships nose left or right depending on the value of degree.
	* The only two vectors we change here are forward and right (and so, only
	* forward, because we calculate right from forward and up vectors)
	*/
	Vector nforward, nup, right, nright;
	float radian;
	float c, s;

	radian=degree*deg2rad;
	c=cos(radian);
	s=sin(radian);

	right = Cross(*up, *forward);

	nforward.x = forward->x*c-right.x*s;
	nforward.y = forward->y*c-right.y*s;
	nforward.z = forward->z*c-right.z*s;

	nright.x = forward->x*s+right.x*c;
	nright.y = forward->y*s+right.y*c;
	nright.z = forward->z*s+right.z*c;

	right=nright;

	//nup=Cross(*forward, right);

	forward->x=nforward.x;
	forward->y=nforward.y;
	forward->z=nforward.z;
	//up->x=nup.x;
	//up->y=nup.y;
	//up->z=nup.z;
}/*}}}*/
void CwCCw(Vector *forward, Vector *up, float degree){/*{{{*/
	/* here we rotate the ship clockwise or counter-clockwise.
	* The only two vectors we change here are up and right (and so, only
	* up, because we calculate right from forward and up vectors)
	*/
	Vector nforward, nup, right, nright;
	float radian;
	float c, s;

	radian=degree*deg2rad;
	c=cos(radian);
	s=sin(radian);

	right = Cross(*up, *forward);

	nright.x = right.x*c-up->x*s;
	nright.y = right.y*c-up->y*s;
	nright.z = right.z*c-up->z*s;

	nup.x = right.x*s+up->x*c;
	nup.y = right.y*s+up->y*c;
	nup.z = right.z*s+up->z*c;

	//nforward=Cross(right, *up);

	right=nright;
	up->x=nup.x;
	up->y=nup.y;
	up->z=nup.z;
	//forward->x=nforward.x;
	//forward->y=nforward.y;
	//forward->z=nforward.z;
}/*}}}*/

/* These are standard GLUT functions. */
void Reshape(int width, int height){ /* {{{ */
	glViewport(0, 0, width, height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	if (width <= height)
		glOrtho(-100.0, 100.0, -100.0*(GLfloat)height/(GLfloat)width, 100.0*(GLfloat)height/(GLfloat)width, -100.0, 100.0);
	else
		glOrtho(-100.0*(GLfloat)width/(GLfloat)height, 100.0*(GLfloat)width/(GLfloat)height, -100.0, 100.0, -100.0, 100.0);
	glMatrixMode(GL_MODELVIEW);

	WindW = width;
	WindH = height;
} /* }}} */
void mouse(int button, int state, int x, int y){ /* {{{ */
	switch (button) {
		case GLUT_LEFT_BUTTON: button = GLI_MOUSE_BUTTON_LEFT; break;
		case GLUT_MIDDLE_BUTTON: button = GLI_MOUSE_BUTTON_MIDDLE; break;
		case GLUT_RIGHT_BUTTON: button = GLI_MOUSE_BUTTON_RIGHT; break;
	}
	switch (state) {
		case GLUT_DOWN: state = GLI_MOUSE_BUTTON_DOWN; break;
		case GLUT_UP: state = GLI_MOUSE_BUTTON_UP; break;
	}
	iface.SendMouseParameters(GLI_MOUSE_CLICK, button, state, x, WindH-y);
} /* }}} */             
void mouse2(int x, int y){ /* {{{ */
	iface.SendMouseParameters(GLI_MOUSE_MOTION_BUTTON, GLI_MOUSE_BUTTON_NONE, GLI_MOUSE_BUTTON_NONE, x, WindH-y);
} /* }}} */
void mouse3(int x, int y){ /* {{{ */
	iface.SendMouseParameters(GLI_MOUSE_MOTION, GLI_MOUSE_BUTTON_NONE, GLI_MOUSE_BUTTON_NONE, x, WindH-y);
} /* }}} */
static void Key(unsigned char key, int x, int y){ /* {{{ */
	if (key == 27){
		exit(0);
	}
	else if(key == '1'){ // cut the engine!
		thrust=0;
	}
	else if(key == '2'){ // reset
		forward.x=0; forward.y=0; forward.z=1;
		up.x=0; up.y=1, up.z=0;
		acceleration.x=0; acceleration.y=0; acceleration.z=0;
		position.x=0; position.y=0; position.z=0;
	}
	else if(key == 'i'){ // pitch down
		UpDown(&forward, &up, 1);
	}
	else if(key == 'k'){ // pitch up
		UpDown(&forward, &up, -1);
	} 
	else if(key == 'l'){ // yaw left
		LeftRight(&forward, &up, 1);
	} 
	else if(key == 'j'){ // yaw right
		LeftRight(&forward, &up, -1);
	} 
	else if(key == 'o'){ // roll clockwise
		CwCCw(&forward, &up, -1);
	} 
	else if(key == 'u'){ // roll counter-clockwise
		CwCCw(&forward, &up, 1);
	} 
	else if(key == 'a'){ // accelerate / forwards thrust
		thrust+=0.1f;
	} 
	else if(key == 'y'){ // break / backwards thrust
		thrust-=0.1f;
	} 
	else{ // oops, unknown key
		fprintf(stderr, "Key: %i\n", key);
		//iface.SendKeyEvent(key, x, WindH-y);
	}
} /* }}} */
void timf(int value){ /* {{{ */
	glutPostRedisplay();
	glutTimerFunc(50, timf, 0);
	
	if (thrust != 0){
		acceleration.x+=forward.x*thrust;
		acceleration.y+=forward.y*thrust;
		acceleration.z+=forward.z*thrust;
	}
	position.x+=acceleration.x;
	position.y+=acceleration.y;
	position.z+=acceleration.z;
	fprintf(stderr, "Position: %2.2f,%2.2f,%2.2f Acceleration: %2.2f,%2.2f,%2.2f\n", position.x, position.y, position.z, acceleration.x, acceleration.y, acceleration.z);
} /* }}} */
void Draw(void){ /* {{{ */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glEnable(GL_TEXTURE_2D);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(60.0f, 800.0f/600.0f, 1.0f, 10000.0f);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	gluLookAt(position.x, position.y, position.z, position.x+forward.x, position.y+forward.y, position.z+forward.z, up.x, up.y, up.z);
	glPopMatrix();
	glDisable(GL_TEXTURE_2D);
	iface.Draw();
	glFlush();              
	glutSwapBuffers();      
} /* }}} */             

int main(int argc, char *argv[]){ // {{{
	float x, y, z;
	int i;

	srand(time(NULL));

	forward.x=0; forward.y=0; forward.z=1;
	up.x=0; up.y=1, up.z=0;
	right=Cross(up, forward);

	position.x=0; position.y=0; position.z=0;
	acceleration=position;
	thrust=0;

	for (i=0; i<5000; i++){
		object = new CGLI_Space();
		object->SetScalingFactor(object->GetID()%5);
		x=(float)(((rand()%50000)/10)*(rand()%2==1?-1:1));
		y=(float)(((rand()%50000)/10)*(rand()%2==1?-1:1));
		z=(float)(((rand()%50000)/10)*(rand()%2==1?-1:1));
		object->SetPosition(x,y,z);
		fprintf(stderr, "New object at %f %f %f\n", x, y, z);
		iface.AddObject(object);
	}

	/* here we're initialising the OpenGL window */
	glutInit(&argc, argv);
	glutInitWindowPosition (0,0);
	WindW=800; WindH=600;

	glutInitWindowSize(WindW, WindH);
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
	(void)glutCreateWindow("Space");

	glutReshapeFunc(Reshape);
	glutDisplayFunc(Draw);
	glutMouseFunc(mouse);
	glutMotionFunc(mouse2);
	glutPassiveMotionFunc(mouse3);
	glutKeyboardFunc(Key);
	glutTimerFunc(50, timf, 0);
	glClearColor(0, 0, 0, 0);
	glColor3f(1, 1, 1);

	//glutFullScreen();
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_TEXTURE_2D);
	glutMainLoop();

	return 0;
} // }}}
